package eu.mhsl.minecraft.pixelblocks.pixelblock;

import eu.mhsl.minecraft.pixelblocks.PixelBlockItem;
import eu.mhsl.minecraft.pixelblocks.PixelBlocksPlugin;
import eu.mhsl.minecraft.pixelblocks.utils.Direction;
import eu.mhsl.minecraft.pixelblocks.utils.MinMaxUtil;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import org.bukkit.*;
import org.bukkit.block.data.BlockData;
import org.bukkit.entity.*;
import org.bukkit.inventory.ItemStack;
import org.bukkit.util.Vector;
import org.jetbrains.annotations.NotNull;

import javax.annotation.Nullable;
import java.io.File;
import java.util.*;

public class PixelBlock {
    private final PixelBlockWorld pixelWorld;

    private final float hitboxOffset = (float) PixelBlocksPlugin.configuration.hitboxOffset();
    private final int pixelsPerBlock = PixelBlocksPlugin.configuration.pixelsPerBlock();

    public Location pixelBlockLocation;
    public Direction facingDirection;
    public List<Pixel> pixels = new ArrayList<>();

    public Interaction hitbox;
    public ItemDisplay barrier;

    public long lastEntryTime;
    public Location lastEntryLocation;
    public UUID ownerUUID;
    public UUID blockUUID;

    public static final int maxPixelsPerBlock = 2000;

    public static @NotNull String getWorldName(@NotNull PixelBlock pixelBlock) {
        return PixelBlocksPlugin.plugin.getDataFolder().getPath() + File.separator + pixelBlock.blockUUID;
    }

    public static @Nullable PixelBlock getPixelBlockFromBlockWorld(World world) {
        return PixelBlocksPlugin.pixelBlocks.stream()
                .filter(block -> block.blockUUID.equals(getUUIDFromWorld(world)))
                .findFirst()
                .orElse(null);
    }

    public static @Nullable UUID getUUIDFromWorld(@NotNull World world) {
        try {
            return UUID.fromString(List.of(world.getName().split("/")).getLast());
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    public static @Nullable PixelBlock getPixelBlockFromPlacedLocation(@NotNull Location placedLocation) {
        Location searchLocation = placedLocation.clone().toBlockLocation();
        searchLocation.setPitch(0);
        searchLocation.setYaw(0);

        return PixelBlocksPlugin.pixelBlocks.stream()
                .filter(block -> Objects.equals(block.pixelBlockLocation, searchLocation))
                .findFirst()
                .orElse(null);
    }

    public PixelBlock(Location originLocation, UUID ownerUUID, UUID blockUUID, Direction direction) {
        PixelBlocksPlugin.pixelBlocks.add(this);

        this.ownerUUID = ownerUUID;
        this.blockUUID = blockUUID;

        this.pixelBlockLocation = originLocation.toBlockLocation();
        this.pixelBlockLocation.setYaw(0);
        this.pixelBlockLocation.setPitch(0);

        this.facingDirection = direction;

        this.pixelWorld = new PixelBlockWorld(this);
    }

    public void enterBlock(@NotNull Player player) {
        if(PixelBlocksPlugin.configuration.onlyEditableByOwner() && !player.getUniqueId().equals(ownerUUID)) {
            player.sendMessage(Component.text("Dieser Pixelblock gehört nicht dir!", NamedTextColor.RED));
            return;
        }

        this.lastEntryLocation = player.getLocation();
        this.lastEntryTime = System.currentTimeMillis();
        PixelBlocksPlugin.database.savePixelBlock(this);

        player.teleport(this.pixelWorld.getSpawnLocation());
    }

    public void setLastEntryLocation(Location lastEntryLocation) {
        this.lastEntryLocation = lastEntryLocation;
    }

    public void spawnInteraction(boolean fullBlock) {
        if(fullBlock) {
            hitbox = (Interaction) pixelBlockLocation.getWorld().spawnEntity(
                    pixelBlockLocation.clone().add(0.5, -hitboxOffset, 0.5),
                    EntityType.INTERACTION
            );
            hitbox.setInteractionHeight(1 + 2*hitboxOffset);
            hitbox.setInteractionWidth(1 + 2*hitboxOffset);
        } else {
            double startingX = MinMaxUtil.getMinProperty(this.pixels, pixel -> pixel.relativeLocation.getX());
            double startingY = MinMaxUtil.getMinProperty(this.pixels, pixel -> pixel.relativeLocation.getY());
            double startingZ = MinMaxUtil.getMinProperty(this.pixels, pixel -> pixel.relativeLocation.getZ());

            double endingX = MinMaxUtil.getMaxProperty(this.pixels, pixel -> pixel.relativeLocation.getX());
            double endingY = MinMaxUtil.getMaxProperty(this.pixels, pixel -> pixel.relativeLocation.getY());
            double endingZ = MinMaxUtil.getMaxProperty(this.pixels, pixel -> pixel.relativeLocation.getZ());

            Location spawnLocation = pixelBlockLocation.clone().add(
                    ((startingX+endingX+1)/2)/pixelsPerBlock,
                    (startingY/pixelsPerBlock)-hitboxOffset,
                    ((startingZ+endingZ+1)/2)/pixelsPerBlock
                );

            float height = (float) (endingY-startingY+1)/pixelsPerBlock + 2*hitboxOffset;

            float width;
            if((endingX-startingX) > (endingZ-startingZ)) {
                width = (float) (endingX-startingX+1)/pixelsPerBlock + 2*hitboxOffset;
            } else {
                width = (float) (endingZ-startingZ+1)/pixelsPerBlock + 2*hitboxOffset;
            }

            if(spawnLocation.getX()+width/2 > this.pixelBlockLocation.getX()+1) {
                spawnLocation.subtract((spawnLocation.getX()+width/2)-(this.pixelBlockLocation.getX()+1), 0, 0);
            }
            if(spawnLocation.getX()-width/2 < this.pixelBlockLocation.getX()) {
                spawnLocation.add(this.pixelBlockLocation.getX()-(spawnLocation.getX()-width/2), 0, 0);
            }

            if(spawnLocation.getZ()+width/2 > this.pixelBlockLocation.getZ()+1) {
                spawnLocation.subtract(0, 0, (spawnLocation.getZ()+width/2)-(this.pixelBlockLocation.getZ()+1));
            }
            if(spawnLocation.getZ()-width/2 < this.pixelBlockLocation.getZ()) {
                spawnLocation.add(0, 0, this.pixelBlockLocation.getZ()-(spawnLocation.getZ()-width/2));
            }

            hitbox = (Interaction) pixelBlockLocation.getWorld().spawnEntity(
                spawnLocation,
                EntityType.INTERACTION
            );

            hitbox.setInteractionHeight(height);
            hitbox.setInteractionWidth(width);
        }
    }

    public void updateEntities() {
        Bukkit.getScheduler().runTask(PixelBlocksPlugin.plugin, () -> {
            this.clearEntities();

            for (int x = 0; x < pixelsPerBlock; x++) {
                for (int y = 0; y < pixelsPerBlock; y++) {
                    for (int z = 0; z < pixelsPerBlock; z++) {
                        Location relativeLocation = new Location(pixelBlockLocation.getWorld(), x, y, z);
                        Location blockLocation = this.pixelWorld.getBuildOrigin();
                        switch (this.facingDirection) {
                            case south -> blockLocation.add(relativeLocation.x(), relativeLocation.y(), relativeLocation.z());
                            case north -> blockLocation.add((pixelsPerBlock-1)-relativeLocation.x(), relativeLocation.y(), (pixelsPerBlock-1)-relativeLocation.z());
                            case east -> blockLocation.add((pixelsPerBlock-1)-relativeLocation.z(), relativeLocation.y(), relativeLocation.x());
                            case west -> blockLocation.add(relativeLocation.z(), relativeLocation.y(), (pixelsPerBlock-1)-relativeLocation.x());
                        }
                        BlockData block = blockLocation.getBlock().getBlockData();

                        if(block.getMaterial() != Material.AIR) {
                            pixels.add(new Pixel(relativeLocation, block, ((double) 1/pixelsPerBlock)));
                        }
                    }
                }
            }

            this.pixels.stream()
                .limit(maxPixelsPerBlock)
                .forEach(pixel -> pixel.place(this.pixelBlockLocation));

            if(this.pixels.size() > maxPixelsPerBlock) {

            }

            if(this.pixels.size() < 5) {
                Location relativeLocation = new Location(pixelBlockLocation.getWorld(), 0, 0, 0);
                BlockData block = Material.GRAY_STAINED_GLASS.createBlockData();
                Pixel newPixel = new Pixel(relativeLocation, block, 1);
                pixels.add(newPixel);
                newPixel.place(this.pixelBlockLocation);

                Location itemDisplayLocation = this.pixelBlockLocation.clone().add(0.5, 0.5, 0.5);
                this.barrier = (ItemDisplay) this.pixelBlockLocation.getWorld().spawnEntity(
                        itemDisplayLocation,
                        EntityType.ITEM_DISPLAY
                );
                this.barrier.setItemStack(ItemStack.of(Material.BARRIER));

                spawnInteraction(true);
            } else {
                spawnInteraction(false);
            }
        });
    }

    public void place(Location placeLocation, Direction direction) {
        Location newLocation = placeLocation.toBlockLocation();
        newLocation.setPitch(0);
        newLocation.setYaw(0);

        @Nullable PixelBlock blockAtLocation = PixelBlock.getPixelBlockFromPlacedLocation(newLocation);
        if(blockAtLocation != null && blockAtLocation != this) throw new IllegalArgumentException("Es können nicht mehrere Pixelblöcke ineinander platziert werden.");

        this.pixelBlockLocation = newLocation;
        this.facingDirection = direction;
        updateEntities();
        PixelBlocksPlugin.database.savePixelBlock(this);
    }

    public void destroy(Player destroyedBy) {
        if(PixelBlocksPlugin.configuration.onlyBreakableByOwner() && !destroyedBy.getUniqueId().equals(ownerUUID)) {
            destroyedBy.sendMessage("Dieser Pixelblock gehört nicht dir!");
            return;
        }

        this.pixelWorld.getPlayersInWorld().forEach(p -> {
            p.sendMessage(Component.text("Der Pixelblock wurde von einem anderen Spieler abgebaut!", NamedTextColor.RED));
            p.teleport(this.lastEntryLocation);
        });

        this.pixelWorld.getEntitiesInWorld().stream()
            .filter(entity -> entity instanceof Item)
            .forEach(entity -> entity.teleport(this.lastEntryLocation));

        this.clearEntities();
        PixelBlocksPlugin.database.deletePixelBlock(this);
        this.pixelBlockLocation.getWorld().playSound(this.pixelBlockLocation, Sound.BLOCK_COPPER_BULB_BREAK, 1.0F, 30);
        this.pixelBlockLocation.getWorld().dropItem(this.pixelBlockLocation.add(new Vector(0.5, 0.5, 0.5)), PixelBlockItem.getBlockAsItem(this));
        this.pixelBlockLocation = null;
    }

    private void clearEntities() {
        Chunk chunk = this.pixelBlockLocation.getChunk();
        if(!chunk.isEntitiesLoaded()) {
            chunk.load(true);
            chunk.getEntities();
        }

        this.pixels.forEach(Pixel::destroy);
        this.pixels.clear();

        if(hitbox != null) {
            this.hitbox.remove();
            this.hitbox = null;
        }

        if(barrier != null) {
            this.barrier.remove();
            this.barrier = null;
        }

        this.pixelBlockLocation.getWorld().getEntities().stream()
                .filter(this::isRelevantEntity)
                .filter(entity -> entity.getLocation()
                        .add(0, hitboxOffset, 0)
                        .toBlockLocation()
                        .equals(this.pixelBlockLocation))
                .forEach(Entity::remove);
    }

    private boolean isRelevantEntity(Entity entity) {
        return entity.getType().equals(EntityType.BLOCK_DISPLAY)
                || entity.getType().equals(EntityType.INTERACTION)
                || entity.getType().equals(EntityType.ITEM_DISPLAY);
    }


    public @NotNull PixelBlockWorld getPixelWorld() {
        return pixelWorld;
    }

    public int getPixelsPerBlock() {
        return pixelsPerBlock;
    }
}
