package eu.mhsl.minecraft.pixelblocks.pixelblock;

import eu.mhsl.minecraft.pixelblocks.PixelBlocksPlugin;
import eu.mhsl.minecraft.pixelblocks.utils.LocationUtil;
import org.bukkit.*;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Player;
import org.bukkit.generator.ChunkGenerator;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.util.List;
import java.util.Objects;
import java.util.Random;

import static eu.mhsl.minecraft.pixelblocks.PixelBlocksPlugin.plugin;

public class PixelBlockWorld {
    private final PixelBlock parentPixelBlock;
    private final World world;

    int worldGrassBorderWidth = 10;
    int pixelsPerBlock = PixelBlocksPlugin.configuration.pixelsPerBlock();

    public static boolean isPixelWorld(@NotNull World world) {
        return world.getName().startsWith(plugin.getDataFolder().getPath());
    }

    public static @NotNull List<World> getOtherWorlds() {
        return Bukkit.getWorlds().stream().filter(w -> !PixelBlockWorld.isPixelWorld(w)).toList();
    }

    public static @NotNull List<World> getPixelBlockWorlds() {
        return Bukkit.getWorlds().stream().filter(PixelBlockWorld::isPixelWorld).toList();
    }

    public PixelBlockWorld(PixelBlock parentPixelBlock) {
        this.parentPixelBlock = parentPixelBlock;
        this.world = this.loadOrCreatePixelWorld();
        this.setBuildingPlatform();
    }

    public boolean allowPlacements(Location blockLocation) {
        Location origin = getBuildOrigin();
        int offset = pixelsPerBlock - 1;
        return blockLocation.x() >= origin.x() && blockLocation.x() <= origin.x() + offset
                && blockLocation.z() >= origin.z() && blockLocation.z() <= origin.z() + offset
                && blockLocation.y() >= origin.y() && blockLocation.y() <= origin.y() + offset;
    }

    public @NotNull String getWorldPathName() {
        return PixelBlocksPlugin.plugin.getDataFolder().getPath() + File.separator + "worlds" + File.separator + this.parentPixelBlock.blockUUID;
    }

    public @NotNull Location getSpawnLocation() {
        return this.getPortalLocation().add(1.5, 1, 2.5);
    }

    public @NotNull Location getPortalLocation() {
        return this.getBuildOrigin().add((double) pixelsPerBlock/2 -2, 0, -worldGrassBorderWidth+3);
    }

    public @NotNull List<Player> getPlayersInWorld() {
        return this.world.getPlayers();
    }

    public @NotNull List<Entity> getEntitiesInWorld() {
        return this.world.getEntities();
    }

    public @NotNull Location getBuildOrigin() {
        return new Location(this.world, 0, -60, 0);
    }

    public @NotNull Location getBuildOriginEnd() {
        return getBuildOrigin().add(pixelsPerBlock, pixelsPerBlock, pixelsPerBlock);
    }

    public @NotNull Location getBorderOrigin() {
        return getBuildOrigin().subtract(1, 1, 1);
    }

    public @NotNull Location getPlatformOrigin() {
        return getBorderOrigin().subtract(worldGrassBorderWidth, 0, worldGrassBorderWidth);
    }

    public @NotNull Location getPlatformOriginEnd() {
        return getBorderOrigin().add(worldGrassBorderWidth + pixelsPerBlock, 0, worldGrassBorderWidth + pixelsPerBlock);
    }

    private World loadOrCreatePixelWorld() {
        final WorldCreator worldCreator = new WorldCreator(getWorldPathName());

        worldCreator.type(WorldType.FLAT);
        worldCreator.generator(new ChunkGenerator() {});

        World world = Bukkit.createWorld(worldCreator);
        Objects.requireNonNull(world);
        world.setGameRule(GameRule.SPAWN_CHUNK_RADIUS, 0);

        world.setGameRule(GameRule.RANDOM_TICK_SPEED, 0);
        world.setGameRule(GameRule.DO_FIRE_TICK, false);
        world.setGameRule(GameRule.DO_MOB_SPAWNING, false);
        world.setGameRule(GameRule.DO_WEATHER_CYCLE, false);
        world.setGameRule(GameRule.DO_VINES_SPREAD, false);
        world.setGameRule(GameRule.DO_DAYLIGHT_CYCLE, false);

        WorldBorder worldBorder = world.getWorldBorder();
        worldBorder.setCenter(getBuildOrigin().add((double) pixelsPerBlock / 2, 0, (double) pixelsPerBlock / 2));
        worldBorder.setSize(pixelsPerBlock + (2*worldGrassBorderWidth));
        worldBorder.setWarningDistance(0);
        worldBorder.setDamageAmount(0);
        return world;
    }

    private void setBuildingPlatform() {
        Bukkit.getScheduler().runTask(PixelBlocksPlugin.plugin, () -> {
            for (int x = 0; x < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; x++) {
                for (int z = 0; z < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; z++) {
                    getPlatformOrigin().add(x, 0, z).getBlock().setType(Material.GRASS_BLOCK);
                }
            }
            for (int x = 0; x < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; x++) {
                for (int z = 0; z < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; z++) {
                    getPlatformOrigin().add(x, -1, z).getBlock().setType(Material.DIRT);
                }
            }

            for (int x = 0; x < (pixelsPerBlock+2); x++) {
                for (int z = 0; z < (pixelsPerBlock+2); z++) {
                    Location currentLocation = getBorderOrigin().add(x, 0, z);

                    if (currentLocation.x() == getBorderOrigin().x() || currentLocation.z() == getBorderOrigin().z()) {
                        currentLocation.getBlock().setType(Material.RED_CONCRETE);
                    } else if (currentLocation.x() == getBorderOrigin().x() + (pixelsPerBlock+1) || currentLocation.z() == getBorderOrigin().z() + (pixelsPerBlock+1)) {
                        currentLocation.getBlock().setType(Material.RED_CONCRETE);
                    }
                }
            }

            Random random = new Random();
            LocationUtil.iterateBlocks(getPlatformOrigin().add(1, 1, 1), getPlatformOriginEnd().add(0, 1, 0), location -> {
                if (allowPlacements(location)) return;
                if (!location.clone().subtract(0, 1, 0).getBlock().getType().equals(Material.GRASS_BLOCK)) return;
                if (!location.getBlock().getType().equals(Material.AIR)) return;

                if (random.nextInt(10) == 0) {
                    Material[] flowers = {
                            Material.DANDELION,
                            Material.POPPY,
                            Material.BLUE_ORCHID,
                            Material.ALLIUM,
                            Material.AZURE_BLUET,
                            Material.RED_TULIP,
                            Material.ORANGE_TULIP,
                            Material.WHITE_TULIP,
                            Material.CORNFLOWER,
                            Material.LILY_OF_THE_VALLEY,
                    };

                    Material randomFlower = flowers[random.nextInt(flowers.length)];
                    location.getBlock().setType(randomFlower);
                }
            });

            for (int x = 0; x < 4; x++) {
                for (int y = 0; y < 5; y++) {
                    getPortalLocation().add(x, y, 0).getBlock().setType(Material.OBSIDIAN);
                }
            }
            for (int x = 1; x < 3; x++) {
                for (int y = 1; y < 4; y++) {
                    getPortalLocation().add(x, y, 0).getBlock().setType(Material.NETHER_PORTAL);
                }
            }
        });
    }
}
